@@ -1,12 +1,12 @@
 #      Disassembler.pm
 #
 #      Copyright (c) 1996 Malcolm Beattie
-#      Copyright (c) 2008,2009 Reini Urban
+#      Copyright (c) 2008,2009,2010,2011,2012 Reini Urban
 #
 #      You may distribute under the terms of either the GNU General Public
 #      License or the Artistic License, as specified in the README file.
 
-$B::Disassembler::VERSION = '1.05_04';
+$B::Disassembler::VERSION = '1.13';
 
 package B::Disassembler::BytecodeStream;
 
@@ -175,30 +175,42 @@ sub GET_IV64 {
   croak "reached EOF while reading I32" unless length($str) == 8;
 
   # Todo: check byteorder
-  return sprintf "0x%09llx", unpack( "q", $str );
+  my $i = unpack( "q", $str );
+  return $i > 8 ? sprintf "0x%09llx", $i : $i;
 }
 
 sub GET_IV {
-
   # Check the header settings, not the current settings.
   $B::Disassembler::ivsize == 4 ? &GET_I32 : &GET_IV64;
-
   # $Config{ivsize} == 4 ? &GET_I32 : &GET_IV64;
 }
 
 sub GET_PADOFFSET {
-
   # Check the header settings, not the current settings.
   $B::Disassembler::ptrsize == 8 ? &GET_IV64 : &GET_U32;
-
   # $Config{ptrsize} == 8 ? &GET_IV64 : &GET_U32;
 }
 
 sub GET_long {
+  # Check the header settings, not the current settings.
+  # B::Disassembler::ivsize or longsize if ge xxx?
+  if ($B::Disassembler::longsize) {
+    return $B::Disassembler::longsize == 8 ? &GET_IV64 : &GET_U32;
+  } else {
+    # return $Config{longsize} == 8 ? &GET_IV64 : &GET_U32;
+    return $B::Disassembler::ivsize == 8 ? &GET_IV64 : &GET_U32;
+  }
+}
 
-  # FIXME: this should check the header settings, not the current settings.
-  # B::Disassembler::ivsize ?
-  $Config{longsize} == 8 ? &GET_IV64 : &GET_U32;
+sub GET_pmflags {
+  my $fh  = shift;
+  my $size = 2;
+  if ($B::Disassembler::blversion ge '"0.07"') {
+    if ($B::Disassembler::perlversion ge '"5.013"') {
+      return $fh->GET_U32;
+    }
+  }
+  return $fh->GET_U16;
 }
 
 package B::Disassembler;
@@ -210,17 +222,50 @@ use strict;
 use B::Asmdata qw(%insn_data @insn_name);
 use Opcode qw(opset_to_ops full_opset);
 use Config qw(%Config);
+use B::Concise;
 
-my $opix;
+BEGIN {
+  if ( $] < 5.009 ) {
+    B::Asmdata->import(qw(@specialsv_name));
+  }
+  else {
+    B->import(qw(@specialsv_name));
+  }
+}
+
+my $ix;
+my $opname;
 our @opname = opset_to_ops(full_opset);
 our (
   $magic,   $archname, $blversion, $ivsize,
   $ptrsize, $longsize, $byteorder, $archflag, $perlversion
 );
+						# >=5.12
+our  @svnames = ("NULL");			# 0
+push @svnames, "BIND"   if $] >= 5.009 and $] < 5.019002; # 1
+push @svnames, ("IV", "NV");			# 2,3
+push @svnames, "RV"     if $] < 5.011;		#
+push @svnames, "PV";
+push @svnames, "INVLIST" if $] >= 5.019002;     # 4
+push @svnames, ("PVIV", "PVNV", "PVMG");	# 4-7
+push @svnames, "BM"     if $] < 5.009;
+push @svnames, "REGEXP" if $] >= 5.011;	# 8
+push @svnames, "GV"     if $] >= 5.009;	# 9
+push @svnames, ("PVLV", "AV", "HV", "CV");	# 10-13
+push @svnames, "GV"     if $] < 5.009;
+push @svnames, ("FM", "IO");			# 14,15
 
 sub dis_header($) {
   my ($fh) = @_;
-  $magic = $fh->GET_U32();
+  my $str = $fh->readn(3);
+  if ($str eq '#! ')  {
+    $str .= $fh->GET_comment_t;
+    $str .= $fh->GET_comment_t;
+    $magic = $fh->GET_U32;
+  } else {
+    $str .= $fh->readn(1);
+    $magic = unpack( "L", $str );
+  }
   warn("bad magic") if $magic != 0x43424c50;
   $archname  = $fh->GET_strconst();
   $blversion = $fh->GET_strconst();
@@ -270,34 +315,76 @@ sub print_insn {
   my ( $insn, $arg, $comment ) = @_;
   undef $comment unless $comment;
   if ( defined($arg) ) {
-    if ( $insn eq 'newopx' or $insn eq 'ldop' ) {    # threaded or unthreaded
+    # threaded or unthreaded
+    if ( $insn eq 'newopx' or $insn eq 'ldop' and $] > 5.007) {
       my $type = $arg >> 7;
       my $size = $arg - ( $type << 7 );
-      $arg .= sprintf( " \t# size:%d, type:%d %s", $size, $type ) if $comment;
-      printf "\n# [%s %d]\n", $opname[$type], $opix++ if $comment;
+      $arg .= sprintf( " \t# size:%d, type:%d %s", $size, $type) if $comment;
+      $opname = $opname[$type];
+      printf "\n# [%s %d]\n", $opname, $ix++;
     }
     elsif ( !$comment ) {
       ;
     }
-    elsif ( $insn eq 'newsvx' ) {
-
-      # TODO which type? SV, AV or GV. check arg
+    elsif ( $insn eq 'comment' ) {
+      $arg .= "comment $arg";
+      $arg .= " \t#" . $comment if $comment ne '1';
+    }
+    elsif ( $insn eq 'stpv' ) {
       $arg .= "\t# " . $comment if $comment ne '1';
-      printf "\n# [%s]\n", 'SV';
+      printf "# -%s- %d\n", 'PV', $ix++;
+    }
+    elsif ( $insn eq 'newsvx' ) {
+      my $type = $arg & 0xff; # SVTYPEMASK
+      $arg .= sprintf("\t# type=%d,flags=0x%x", $type, $arg);
+      $arg .= $comment if $comment ne '1';
+      printf "\n# [%s %d]\n", $svnames[$type], $ix++;
+    }
+    elsif ( $insn eq 'newpadlx' ) {
+      $arg .= "\t# " . $comment if $comment ne '1';
+      printf "\n# [%s %d]\n", "PADLIST", $ix++;
+    }
+    elsif ( $insn eq 'newpadnlx' ) {
+      $arg .= "\t# " . $comment if $comment ne '1';
+      printf "\n# [%s %d]\n", "PADNAMELIST", $ix++;
+    }
+    elsif ( $insn eq 'newpadnx' ) {
+      $arg .= "\t# " . $comment if $comment ne '1';
+      printf "\n# [%s %d]\n", "PADNAME", $ix++;
     }
     elsif ( $insn eq 'gv_stashpvx' ) {
       $arg .= "\t# " . $comment if $comment ne '1';
-      printf "\n# [%s]\n", "STASH";
+      printf "\n# [%s %d]\n", "STASH", $ix++;
+    }
+    elsif ( $insn eq 'ldspecsvx' ) {
+      $arg .= "\t# $specialsv_name[$arg]";
+      $arg .= $comment if $comment ne '1';
+      printf "\n# [%s %d]\n", "SPECIAL", $ix++;
     }
     elsif ( $insn eq 'ldsv' ) {
-
-      # TODO which type? SV, AV or GV. check arg
       $arg .= "\t# " . $comment if $comment ne '1';
-      printf "\n# -%s-\n", 'SV';
+      printf "# -%s-\n", 'GP/AV/HV/NULL/MG';
     }
     elsif ( $insn eq 'gv_fetchpvx' ) {
       $arg .= "\t# " . $comment if $comment ne '1';
-      printf "\n# [%s]\n", 'prototype';
+      printf "\n# [%s %d]\n", 'GV', $ix++;
+    }
+    elsif ( $insn eq 'sv_magic' ) {
+      $arg .= sprintf( "\t# '%s'", chr($arg) );
+    }
+    elsif ( $insn =~ /_flags/ ) {
+      my $f = $arg;
+      $arg .= sprintf( "\t# 0x%x", $f ) if $comment;
+      $arg .= " ".B::Concise::op_flags($f) if $insn eq 'op_flags' and $comment;
+    }
+    elsif ( $comment and $insn eq 'op_private' ) {
+      my $f = $arg;
+      $arg .= sprintf( "\t# 0x%x", $f );
+      $arg .= " ".B::Concise::private_flags($opname, $f);
+    }
+    elsif ( $insn eq 'op_type' and $] < 5.007 ) {
+      my $type = $arg;
+      $arg .= sprintf( "\t# [ %s ]", $opname[$type] );
     }
     else {
       $arg .= "\t# " . $comment if $comment ne '1';
@@ -325,7 +412,7 @@ sub disassemble_fh {
   my $out     = shift;
   my $verbose = shift;
   my ( $c, $getmeth, $insn, $arg );
-  $opix = 1;
+  $ix = 1;
   bless $fh, "B::Disassembler::BytecodeStream";
   dis_header($fh);
   if ($verbose) {
@@ -348,6 +435,7 @@ sub disassemble_fh {
       warn "Illegal instruction code $c at stream offset $pos.\n";
     }
     $getmeth = $insn_data{$insn}->[2];
+    #warn "EOF at $insn $getmeth" if $fh->eof();
     $arg     = $fh->$getmeth();
     if ( defined($arg) ) {
       &$out( $insn, $arg, $verbose );
@@ -435,22 +523,31 @@ B<longsize> is $Config{longsize} of the assembling perl.
 A number, 4 or 8.
 Only since blversion 0.06_03.
 
-B<byteorder> is a string of "0x12345678" on big-endian or "0x56781234" (?)
+B<byteorder> is a string of "0x12345678" on big-endian or "0x78563412" (?)
 on little-endian machines. The beginning "0x" is stripped for compatibility
 with intermediate ByteLoader versions, i.e. 5.6.1 to 5.8.0,
 Added with blversion 0.06_03, and also with blversion 0.04.
 See L<BcVersions>
 
 B<archflag> is a bitmask of opcode platform-dependencies.
-Currently used is only bit 1 for USE_ITHREADS.
+Currently used:
+  bit 1 for USE_ITHREADS
+  bit 2 for MULTIPLICITY
 Added with  blversion 0.06_05.
 
-B<perlversion> $] of the perl, which produced this bytecode, as string.
+B<perlversion> $] of the perl which produced this bytecode as string.
 Added with blversion 0.06_06.
 
 =head1 AUTHORS
 
-Malcolm Beattie, C<mbeattie@sable.ox.ac.uk>.
-Reini Urban <rurban@cpan.org> since 2008.
+Malcolm Beattie C<MICB at cpan.org> I<(retired)>,
+Reini Urban C<perl-compiler@googlegroups.com> since 2008.
 
 =cut
+
+# Local Variables:
+#   mode: cperl
+#   cperl-indent-level: 2
+#   fill-column: 100
+# End:
+# vim: expandtab shiftwidth=2:
